----------------------------------------------------------------------------------

-- LGP-30 drum computer simulation
-- Top-level module
--
-- jm 26.7.2016
-- jm 30.7.2017 USB port added
--
--
-- Clocking considerations:
--
-- Original LGP-30 has a state-sensitive clocking, with short clock spikes going low (T) or high (Tn).
-- Flip-flops are triggered by T or Tn (different for logic and read flip-flops), but all at the same time.
-- This replica design uses edge-sensitive clocking (on rising edge) and a single clock signal instead.
--
-- A 10 MHz I/O clock is used to drive all peripherals (control panel, serial tty, lcd) 
-- and define their timing towards the outside world. 
-- This clock is also used as SCL for the LCD SPI, hence may not be faster than 20 MHz.
--
-- The LGP-30 clock is asymmetrical, high for only one tick of the I/O clock.
-- This is used by the peripheral modules to run delay and pulse timers at the LGP-30 clock rate.
-- (Peripherals are clocked by the I/O clock, but use the LPG-30 clock to gate timer ticks.)
--
--
-- Naming conventions:
--
-- LGP-30 signal names used are in line with the German service manual (Schoppe & Faeser).
-- Suffixes used:
-- 'p' = prime, ' superscript
-- 's' = star, * superscript
-- 'n' = not, _ underline
--
-- Bits in the simulated LGP-30 are numbered according to the LGP-30 documentation: 
-- Bit numbering starts with 1, and bit 1 is the *most* significant bit!
-- (Vectors indexed "1 to 6" are used to represent these in VHDL.)
-- But the *least* significant bit appears first in time during processing (as needed for carries).
--
-- Bit vectors in non-LGP-30 modules are numbered X downto 0, with bit 0 the least significant.
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

library work;
use work.constants.all;

entity LGP30 is
	port(
		-- system clock
		clk100: 	in std_logic;
		
		-- panel switches and LEDs
		switch: in std_logic_vector (24 downto 0);
		led: out std_logic_vector (22 downto 0);

		-- external serial I/O
		serRx: in std_logic;
		serTx: out std_logic;
		serRTS: out std_logic;					-- '0' = we are ready to receive data
		serCTS: in std_logic;					-- '0' = PC is ready to receive data - used during Xmodem transfers only
		
		-- HDMI data
		hdmi_out_p: out STD_LOGIC_VECTOR (3 downto 0);
		hdmi_out_n: out STD_LOGIC_VECTOR (3 downto 0);
		hotplug: in std_logic := '0';			-- not used for now

		-- LCD (Electronic Assembly EA DOGM132, 32*132 pixel, SPI bus)
		lcd_si: out std_logic;  				-- data
		lcd_scl: out std_logic;					-- clock 
		lcd_csn: out std_logic;					-- chip select, active low
		lcd_a0: out std_logic;					-- high = data, low = command
		lcd_resn: out std_logic;				-- reset, active low

		-- external USB bus
		usb_dp : inout  STD_LOGIC;				-- Data+
		usb_dn : inout  STD_LOGIC;				-- Data-
		
		-- USB/UART status LEDs (for troubleshooting only)
		usb_led: out STD_LOGIC;					-- status LED (USB host connection)
		usb_act_led: out STD_LOGIC;			-- USB serial in use (character received)
		uart_act_led: out STD_LOGIC;			-- UART in use (character received)

		-- relay output (clicker) on landscape board
		relay: out std_logic;

		-- debug outputs
		test: inout std_logic_vector (7 downto 0)
		
--		Dedicated external scope outputs are no longer present.
--		Use the test/expansion connector if these are desired, with resistor network on an external adapter.
--		See test connector wiring and comments at the bottom of this file.
--		trace: out std_logic;								-- trace - alternates between A, R, C
--		offset: out std_logic_vector (1 downto 0);	-- vertical offsets  alternate between 00, 01, 11
--		trig: out std_logic;									-- trigger 
	);
end LGP30;

architecture Behavioral of LGP30 is

------------------------------------------------
-- component declarations
------------------------------------------------

	component clock 
		port ( 
			clk_ext: 	in  std_logic;				-- external clock input
			clk_usb:		out std_logic;				-- USB PHY clock
			clk_io: 		buffer std_logic;			-- I/O clock, 10 MHz
			clk_lgp30:	out std_logic;				-- LGP-30 CPU clock. High lasts one I/O clock cycle
			clk_lgp30_unbuf: out std_logic;		-- same, but unbuffered - for use in io clock and lgp clk synchronization in panel and flexowriter modules

			-- user control for LGP-30 clock rate
			cup, cdn:	in std_logic;				-- pulse high for one clk_io cycle
			cstep:		in std_logic;				-- set high to force clock speed to single step (XModem mode)
			cpow: 		out freq_index				-- slowing clock down by 2^cpow
		);
	end component;

	component logic
		port (
		
			-- system clock
			clk: in std_logic;						-- LGP-30 system clock
			clk_en: in std_logic;					-- clock enable (disabled during xmodem transfers)
			
			-- main memory
			V, Vn: in std_logic;						-- read head
			Vp: out std_logic;						-- write head
			W: buffer std_logic;						-- write enable
			P: out std_logic_vector (1 to 6);	-- track selector
		
			-- recycling registers
			A, An: in std_logic;						-- accumulator read, 32 bits behind write
			As, Asn: in std_logic;					-- accumulator read, 65 bits behind write
			Ap: out std_logic;						-- accumulator write, awlays enabled

			R, Rn: in std_logic;						-- instruction read, 32 bits behind write
			Rp: out std_logic;						-- instruction write, always enabled

			C, Cn: in std_logic;						-- program counter read, 32 bits behind write
			Cp: out std_logic;						-- program counter write, always enabled
			
			-- timing signals
			S1, S1n: in std_logic;
			S2, S2n: in std_logic;
			S3, S3n: in std_logic;

			-- control panel, input and status output
			bs: in std_logic;							-- start
			bq, bqn: in std_logic;					-- manual input
			bcn: in std_logic;						-- clear counter
			br, brn: in std_logic;					-- fill instruction
			Ps, Psn: in std_logic;					-- execute instruction
			O1, O1n: in std_logic;					-- single step

			T: in std_logic_vector (1 to 4);		-- breakpoint switches
			D, Dn: in std_logic;						-- select 6 bit input
			Tc: in std_logic;							-- transfer control
			
			Tbs: out std_logic;						-- "blocked state" flag - drives stop/compute lights

			-- flexowriter
			Tp: out std_logic_vector (1 to 6);	-- character to be printed/punched
			PPs, PPsn: in std_logic_vector (1 to 6);	-- character from keyboard or tape reader
			Tx: out std_logic;						-- trigger printing
			Tread: out std_logic;					-- trigger reading of next 4/6 bit input
			Tdelay: out std_logic;					-- trigger delay in Flexwriter (ignored in simulation?)
															-- note: Tstart is received via bs input
			-- oscillosope display: trigger
			t3, t3n: buffer std_logic;				-- high during sign bit cycle
			
			-- flip-flop outputs to LCD
			Fout, Gout, Hout, Kout, Lout: out std_logic; -- phase control
			Qout: out std_logic_vector (1 to 4);	-- instruction decoding
			
			-- debug outputs
			test: out std_logic_vector (7 downto 0)
		);
	end component;
	
	component drum
		port (
			
			-- system clock
			clk: in std_logic;						-- LGP-30 system clock
			
			-- main memory
			V, Vn: out std_logic;					-- read head
			Vp: in std_logic;							-- write head
			W: in std_logic;							-- write enable
			P: in std_logic_vector (1 to 6);		-- track selector
		
			-- recycling registers
			A, An: out std_logic;					-- accumulator read, 32 bits behind write
			As, Asn: out std_logic;					-- accumulator read, 65 bits behind write
			Ap: in std_logic;							-- accumulator write, always enabled

			R, Rn: out std_logic;					-- instruction read, 32 bits behind write
			Rp: in std_logic;							-- instruction write, always enabled

			C, Cn: out std_logic;					-- program counter read, 32 bits behind write
			Cp: in std_logic;							-- program counter write, always enabled
			
			-- timing signals
			S1, S1n: out std_logic;
			S2, S2n: out std_logic;
			S3, S3n: out std_logic;
			
			-- memory access for HDMI display
			vid_clk: in std_logic;
			vid_adr: in std_logic_vector (16 downto 0);
			vid_data: out std_logic;
			vid_regs: out std_logic_vector (5 downto 0);	-- S3, S2, S1, A, R, C
			drumpos: out std_logic_vector (10 downto 0);	-- current drum rotary position

			-- control for xmodem-driven memory access
			sec_reset: in std_logic								-- reset rotary position to sector 0
			
		);
	end component;
	
	component panel
		port (	

			-- clocks
			io_clk: in std_logic;						-- I/O clock; for debounce timer only
			lgp_clk: in std_logic;						-- LGP-30 bit clock; determines timing towards LGP-30 logic unit
					
			-- inside simulated LGP-30 
			bs: out std_logic;							-- start
			bq, bqn: out std_logic;						-- manual input
			bcn: out std_logic;							-- clear counter
			br, brn: out std_logic;						-- fill instruction
			Ps, Psn: out std_logic;						-- execute instruction
			O1, O1n: out std_logic;						-- single step

			T: out std_logic_vector (1 to 4);		-- breakpoint switches
			D, Dn: out std_logic;						-- select 6 bit input
			Tc: out std_logic;							-- transfer control
			
			Tbs: in std_logic;							-- "blocked state" flag - drives stop/compute lights
			
			-- buttons on FlexoWriter in the real LGP-30
			readActive: in std_logic;					-- signals that FlexoWriter is ready to read (for read indicator led)
			readToggle: out std_logic;					-- alternate between ""start read" and "stop read"; actual status controlled by Flexowriter
			connected: out std_logic;					-- Flexowriter can send start signals to the logic unit
				
			-- Encoder for simulator control
			EncUp, EncDn: out std_logic;				-- pulses high for one io_clk
			EncPushed: out std_logic;					-- direct read of pushbutton;
			EncSelect: out std_logic;					-- toggled by encoder pushbutton
		
			-- option switches (toggles) on Mimas board
			option: out std_logic_vector (3 downto 0);
		
			-- external switches and LEDs
			switch: in std_logic_vector (24 downto 0);
			led: out std_logic_vector (22 downto 0)
			
		);
	end component;

	component flexowriter 
		port (
		
			-- inside simulated LGP-30
			lgp_clk: in std_logic;							-- LGP-30 bit clock; determines timing towards LGP-30 logic unit
			Tp: in std_logic_vector (1 to 6);			-- character to be printed/punched
			PPs, PPsn: out std_logic_vector (1 to 6);	-- character from keyboard or tape reader
			Ps, Psn: out std_logic;							-- signal: character received 
			Ttrig: in std_logic;								-- trigger printing
			Tread: in std_logic;								-- trigger reading of next 4/6 bit input
			Tdelay: in std_logic;							-- trigger delay in Flexwriter (ignored in simulation)
			Tstart: out std_logic;							-- received Cond Stop code; restart logic unit from blocked state

			-- connection to UART unit
			io_clk: in std_logic;							-- I/O clock
			uart_data_tx: out std_logic_vector(7 downto 0);		
			uart_data_rx: in std_logic_vector(7 downto 0);		
			write_to_uart_tx: out std_logic;				-- trigger 1 byte write to Tx buffer
			read_from_uart_rx: out std_logic;			-- trigger 1 byte read from Rx buffer
			uart_rx_half_full: in std_logic; 			-- Rx buffer >= half full?
			uart_rx_data_present: in std_logic; 		-- Rx buffer has character?

			-- enable output of ESC sequences
			use_ESC: in std_logic;
		
			-- read start/stop control: connections to control panel
			readActive: out std_logic;						-- signals that FlexoWriter is ready to read (for read indicator led)
			readToggle: in std_logic;						-- alternate between ""start read" and "stop read"; actual status controlled by Flexowriter
		
			-- handshake with xmodem unit
			xmodem_active: buffer std_logic;				-- xmodem transfer is active
			xmodem_done: in std_logic;						-- strobes high when xmodem is done (or has failed)
		
			-- simulate printing noise (via external relais)
			noise: out std_logic;

			-- debug outputs
			test: out std_logic_vector (7 downto 0)
		);
	end component;
	
	component crt
		port (
		
			-- all registers are presented during each word cycle
			A: in std_logic;			-- accumulator to display
			R: in std_logic;			-- instruction register to display
			C: in std_logic;			-- program counter to display
			
			-- trigger
			-- triggers horizontal deflection (right to left!) and 3-step counter for vertical offset
			T: in std_logic;			-- bit clock
			t3: in std_logic;			-- high during sign bit cycle
			
			-- output to external scope
			trace: out std_logic;	-- trace - alternates between A, R, C
			offset: buffer std_logic_vector (1 downto 0);	-- vertical offsets  alternate between 00, 01, 11
			trig: out std_logic		-- trigger 
			
		);
	end component;

   component HDMI
		port (
			-- system clock
			clk100: in std_logic;
		
			-- data input(main drum memory)
			clk: out std_logic;
			adr: buffer std_logic_vector (16 downto 0);
			data: in std_logic;
			regs: in std_logic_vector (5 downto 0);		-- S3, S2, S1, A, R, C

			-- data for display of r/w heads (used at slow clock speeds only)
			drumpos: in std_logic_vector (10 downto 0);	-- current drum rotary position
			track: in std_logic_vector (5 downto 0);		-- current LGP-30 track address
			write_en: in std_logic;								-- write enable for LGP-30 heads

			-- flip-flops for status display
			F, G, H, K, L: in std_logic;
			P: in std_logic_vector (1 to 6);
			Q: in std_logic_vector (1 to 4);
			
			-- display mode control
			io_clk: in std_logic;
			right: in std_logic;				-- pulses high to move display right (encoder controlled)
			left: in std_logic;				-- pulses high to move display left (encoder controlled)
			animate: in std_logic;			-- animate the display (at slow clock speeds)
			showff: in std_logic;			-- show flip-flops
			
			-- HDMI output
			hdmi_p : out std_logic_vector(3 downto 0);
			hdmi_n : out std_logic_vector(3 downto 0);

			-- debug outputs
			test: out std_logic_vector (7 downto 0)
		);
	end component;
	
	component lcd
		port (
			-- all registers are presented during each word cycle
			A: in std_logic;					-- accumulator to display
			R: in std_logic;					-- instruction register to display
			C: in std_logic;					-- program counter to display
			
			-- bit clock and trigger
			T: in std_logic;					-- bit clock
			t3: in std_logic;					-- high during sign bit cycle
			io_clk: in std_logic;			-- 10 MHz

			-- display settings 
			ext_disp: in std_logic;			-- extended display on?
			cont_up: in std_logic;			-- increase contrast value (1 step)
			cont_dn: in std_logic;			-- decrease contrast value (1 step)

			-- flip-flops for status display
			F, G, H, K, L: in std_logic;
			P: in std_logic_vector (1 to 6);
			Q: in std_logic_vector (1 to 4);
			
			-- clock rate display
			inv_rate: in std_logic;			-- invert clock rate display?
			cpow: in freq_index; 			-- slowing clock down by 2^cpow; 0 = 125 kHz
			
			-- output to external display.
			-- Electronic Assembly EA DOGM132, 32*132 pixel, ST7565R display controller, SPI interface.
			si: out std_logic;  				-- data
			scl: out std_logic;				-- clock
			cs_n: out std_logic;				-- chip select, active low
			a0: out std_logic;				-- high = data, low = command
			res_n: out std_logic				-- reset, active low
		);
	end component;

	component usb_uart_mux
		port (
			-- clocks 
			io_clk : in std_logic;						-- IO clock, 10 MHz
			usb_clk : in std_logic;						-- USB PHY clock, 60 MHz

			-- status of serial interfaces
			usb_active: out std_logic;					-- USB data received, interface in use
			uart_active: out std_logic;				-- UART data received, interface in use

			-- communication with internal units
			-- these are synchronized to io_clk
			data_tx:in std_logic_vector(7 downto 0);		
			data_rx: out std_logic_vector(7 downto 0);		
			write_strobe: in std_logic;				-- trigger 1 byte write to Tx buffer
			read_strobe: in std_logic;					-- trigger 1 byte read from Rx buffer
			tx_half_full: out std_logic; 				-- Tx buffer >= half full?
			rx_half_full: out std_logic; 				-- Rx buffer >= half full?
			rx_data_present: out std_logic; 			-- Rx buffer has character?
			
			-- external USB bus
			usb_dp : inout  STD_LOGIC;					-- Data+
			usb_dn : inout  STD_LOGIC;					-- Data-
			usb_led: out STD_LOGIC;						-- status LED
			
			-- external serial I/O
			uart_Rx: in std_logic;
			uart_Tx: out std_logic;
			uart_RTS: out std_logic;
			uart_CTS: in std_logic
		);
	end component;
	
	component xmodem 
		port (
			-- begin/end xmodem mode
			xmodem_active: in std_logic;					-- xmodem transfer is active
			xmodem_done: out std_logic;					-- strobe high when we are done (or have failed)
		
			-- memory access to LGP-30 drum 
			-- system clock
			clk_tick: out std_logic;						-- can control clock from this unit via the single step mode

			-- main memory
			V: in std_logic;									-- read head
			Vp: out std_logic;								-- write head
			W: out std_logic;									-- write enable
			P: out std_logic_vector (1 to 6);			-- track selector
			sec_reset: out std_logic;						-- reset rotary position to sector 0
			
			-- registers
			A, R, C: in std_logic;							-- register read heads
			Ap, Rp, Cp: out std_logic;						-- register write heads - always on!
			
			-- connection to UART unit
			io_clk: in std_logic;							-- I/O clock
			uart_data_tx: out std_logic_vector(7 downto 0);		
			uart_data_rx: in std_logic_vector(7 downto 0);		
			write_to_uart_tx: out std_logic;				-- trigger 1 byte write to Tx buffer
			read_from_uart_rx: out std_logic;			-- trigger 1 byte read from Rx buffer
			uart_tx_half_full: in std_logic; 			-- Tx buffer >= half full?
			uart_rx_half_full: in std_logic; 			-- Rx buffer >= half full?
			uart_rx_data_present: in std_logic; 		-- Rx buffer has character?

			-- debug outputs
			test: out std_logic_vector (7 downto 0)
		);
	end component;

------------------------------------------------
-- global "wiring"
------------------------------------------------

-- buffered clock
signal clk100_buf: std_logic;

-- clock for simulated lgp-30
-- different from real LGP-30, this is not coming from the magnet drum module,
-- but generated by a separate clock divider
signal clk, clk_unbuf: std_logic;
signal cpow: freq_index;			-- slowing clock down by 2^cpow

-- I/O clock for panel (debouncing) and flexowriter (UART)
signal ioclk: std_logic;

-- USB PHY layer clock
signal usbclk: std_logic;
		
-- magnetic drum memory
signal A, An, As, Asn, Ap: std_logic;
signal R, Rn, Rp: std_logic;
signal C, Cn, Cp: std_logic;
signal V, Vn, Vp, W: std_logic;
signal P: std_logic_vector (1 to 6);
signal S1, S1n, S2, S2n, S3, S3n: std_logic;

-- control panel
signal bs, bq, bqn, bcn, br, brn: std_logic;
signal Ps, Psn, O1, O1n, D, Dn, Tc: std_logic;
signal T: std_logic_vector (1 to 4);
signal Tbs: std_logic;
signal connected: std_logic;									-- start CPU signal from flexowriter is valid
signal option: std_logic_vector (3 downto 0);			-- option switches

-- flexowriter
signal tty_write: std_logic_vector (1 to 6);		-- character to be printed/punched
signal tty_read, tty_readn: std_logic_vector (1 to 6);		-- character from keyboard or tape reader
signal tty_trigger: std_logic;								-- trigger printing
signal tty_doread: std_logic;									-- trigger reading of next 4/6 bit input
signal tty_delay: std_logic;									-- trigger delay in Flexwriter (not needed in simulation?)
signal tty_start: std_logic;									-- received Cond Stop code; restart logic unit from blocked state
signal tty_Ps, tty_Psn: std_logic;							-- received a character, signal to logic unit
signal tty_reading: std_logic;								-- tape reader active (not RTS in simulation)
signal tty_toggle: std_logic;									-- start/stop read button (from control panel)
signal tty_noise: std_logic;									-- simulated typewriter noise (to external relay)

-- oscilloscope
signal t3, t3n: std_logic;										-- internal trigger from logic unit
signal trace: std_logic;										-- trace - alternates between A, R, C
signal offset: std_logic_vector (1 downto 0);			-- vertical offsets  alternate between 00, 01, 11
signal fat_trace: std_logic := '0';							-- 5 Mhz to fatten trace
signal trig: std_logic;											-- trigger 

-- HDMI
signal vid_clk: std_logic;										-- video clock (generated in HDMI module)
signal vid_adr: std_logic_vector (16 downto 0);			-- address to drum memory (6 bit track, 6 bit sector, 5 bit inside wordclo)
signal vid_data: std_logic;									-- one bit data from drum memory
signal vid_drumpos: std_logic_vector (10 downto 0);	-- current rotary position of drum
signal vid_regs: std_logic_vector (5 downto 0);			-- S3, S2, S1, A, R, C

-- Encoder
signal enc_up, enc_dn: std_logic;							-- pulse high for 1 io_clk cycle to indicate encoder moves
signal enc_but: std_logic;										-- encoder pushbutton
signal enc_sel: std_logic;										-- status of the encoder pushbutton toggle
	
-- to LCD for flip-flop display 
signal F, G, H, K, L: std_logic;								-- phase control
signal Q: std_logic_vector (1 to 4);						-- instruction decoding

-- xmodem transfers
signal xmodem_act: std_logic;									-- flexowriter requests xmodem transfer
signal xmodem_done: std_logic;								-- xmodem signals completion to flexowriter
signal drum_sec_reset: std_logic;							-- xmodem resets drum position to read/write a track
signal clk_tick: std_logic;									-- individual clock ticks

signal Vp_xmodem: std_logic;									-- write head
signal W_xmodem: std_logic;		 							-- write enable
signal Ap_xmodem: std_logic;									-- A register write head
signal Rp_xmodem: std_logic;									-- R register write head
signal Cp_xmodem: std_logic;									-- C register write head
signal P_xmodem: std_logic_vector (1 to 6);				-- track selector
signal P_mux: std_logic_vector (1 to 6);					-- track selector (mux'd between xmodem and CPU output)

-- UART
signal uart_data_tx: std_logic_vector(7 downto 0);
signal write_to_uart_tx: std_logic;
signal uart_data_rx: std_logic_vector(7 downto 0);
signal read_from_uart_rx: std_logic;
signal uart_rx_data_present: std_logic;
signal uart_rx_half_full: std_logic;
signal uart_tx_half_full: std_logic;

signal uart_data_tx_xmodem: std_logic_vector(7 downto 0);-- tx data from xmodem unit, mux'd into uart_data_tx_mux 
signal uart_data_tx_mux: std_logic_vector(7 downto 0);	-- mux'd data between xmodem and flexowriter output
signal write_to_uart_tx_xmodem: std_logic;					-- write strobe from xmodem unit
signal read_from_uart_rx_xmodem: std_logic;					-- read strobe fro xmodem unit

signal usb_active: std_logic;										-- USB data received, interface in use
signal uart_active: std_logic;									-- UART data received, interface in use


begin

------------------------------------------------
-- component instances
------------------------------------------------

	-- External clock buffering
	-- Must use a BUFG here to enable routing to both PLLs (hdmi and io/usb clocks).
	--	clk_buf : IBUFG port map (O => clk100_buf, I => clk100);		-- not routable
	clk_buf : BUFG port map (O => clk100_buf, I => clk100);
	 
	inst_clock: clock
		PORT MAP (
			clk_ext		=> clk100_buf,
			clk_usb		=> usbclk,
			clk_io 		=> ioclk,
			clk_lgp30	=> clk,
			clk_lgp30_unbuf => clk_unbuf,
			cup			=> (enc_up and enc_sel and not enc_but and not xmodem_act),	-- disable during xmodem mode
			cdn			=> (enc_dn and enc_sel and not enc_but and not xmodem_act) or (clk_tick and xmodem_act),	-- xmodem can trigger individual clock steps
			cstep			=> xmodem_act,							-- force single step during xmodem mode
			cpow			=> cpow
		);

	inst_logic: logic
		PORT MAP (
			-- system clock
			clk 		=> clk,	
			clk_en	=> not xmodem_act,		-- stop cpu clock while xmodem is active 
			
			-- main memory
			V			=> V,
			Vn			=> Vn,
			Vp			=> Vp,
			W		 	=> W,
			P		 	=> P,
			
			-- recycling registers
			A		 	=> A,
			An			=> An,
			As		 	=> As,
			Asn		=> Asn,
			Ap		 	=> Ap,	
			R		 	=> R,
			Rn			=> Rn,
			Rp		 	=> Rp,
			C		 	=> C,
			Cn		 	=> Cn,
			Cp		 	=> Cp,
			
			-- timing signals
			S1 		=> S1,
			S1n 		=> S1n,
			S2 		=> S2,
			S2n 		=> S2n,
			S3 		=> S3,
			S3n 		=> S3n,
			
			-- control panel input
			bs 		=> bs OR (connected AND tty_start),	-- control panel and Flexowriter both provide start signals
			bq 		=> bq,
			bqn 		=> bqn,
			bcn 		=> bcn,
			br 		=> br,
			brn 		=> brn,
			Ps 		=> Ps or tty_Ps,
			Psn 		=> Psn and tty_Psn,
			O1 		=> O1,			
			O1n 		=> O1n,			
			T 			=> T,
			D 			=> D,
			Dn 		=> Dn,
			Tc 		=> Tc,
			Tbs		=> Tbs,
			
			-- flexowriter
			Tp 		=> tty_write,
			PPs 		=> tty_read,
			PPsn 		=> tty_readn,
			Tx 		=> tty_trigger,
			Tread		=> tty_doread,
			Tdelay	=> tty_delay,

			-- oscillosope display: trigger
			t3		 		=> t3,
			t3n			=> t3n,
			
			-- LCD
			Fout			=> F,
			Gout			=> G,
			Hout			=> H,
			Kout			=> K,
			Lout			=> L,
			Qout			=> Q,
			
			-- debug outputs
			test			=> open
		);

	inst_drum: drum
		PORT MAP (
			-- system clock
			clk 		=> clk,

			-- main memory
			V			=> V,
			Vn			=> Vn,
			Vp			=> (Vp and not xmodem_act) or (Vp_xmodem and xmodem_act),
			W		 	=> (W  and not xmodem_act) or (W_xmodem  and xmodem_act),
			P		 	=> P_mux,
			
			-- recycling registers
			A		 	=> A,
			An			=> An,
			As		 	=> As,
			Asn		=> Asn,
			Ap		 	=> (Ap and not xmodem_act) or (Ap_xmodem and xmodem_act),	
			R		 	=> R,
			Rn			=> Rn,
			Rp		 	=> (Rp and not xmodem_act) or (Rp_xmodem and xmodem_act),
			C		 	=> C,
			Cn		 	=> Cn,
			Cp		 	=> (Cp and not xmodem_act) or (Cp_xmodem and xmodem_act),
			
			-- timing signals
			S1 		=> S1,
			S1n 		=> S1n,
			S2 		=> S2,
			S2n 		=> S2n,
			S3 		=> S3,
			S3n 		=> S3n,
			
			-- memory access for HDMI display
			-- only main drum memory for now
			vid_clk => vid_clk,
			vid_adr => vid_adr,
			vid_data => vid_data,
			vid_regs	=> vid_regs,
			drumpos => vid_drumpos,
			sec_reset => drum_sec_reset
		);
		
	inst_panel: panel
		PORT MAP (
		
			io_clk	=> ioclk,				-- I/O clock; for debounce timer only
			lgp_clk	=> clk_unbuf,			-- CPU clock: determines pulse lengths to logic unit
			
			-- control panel input buttons
			bs 		=> bs,
			bq 		=> bq,
			bqn 		=> bqn,
			bcn 		=> bcn,
			br 		=> br,
			brn 		=> brn,
			Ps 		=> Ps,
			Psn 		=> Psn,
			O1 		=> O1,			
			O1n 		=> O1n,			
			T 			=> T,
			D 			=> D,
			Dn 		=> Dn,
			Tc 		=> Tc,
			Tbs		=> Tbs,
			
			-- flexowriter buttons, in control panel unit here
			readActive=> tty_reading,
			readToggle => tty_toggle,
			connected => connected,
			
			-- encoder outputs
			EncUp		=> enc_up,
			EncDn		=> enc_dn,
			EncPushed => enc_but,
			EncSelect=> enc_sel,
			
			option	=> option,
			
			-- external switches and leds
			switch	=> switch,
			led		=> led
		);
	
	inst_crt: crt
		PORT MAP (
			-- system clock and sign bit trigger
			T	 		=> clk,
			t3 		=> t3,
			
			-- recycling registers
			A		 	=> A,
			R			=> R,
			C			=> C,	
			
			-- output to external scope
			trace		=> trace,
			offset	=> offset,
			trig		=> trig
		);

	inst_flexowriter: flexowriter
		PORT MAP (
			-- inside simulated LGP-30
			lgp_clk	=> clk_unbuf,					-- CPU clock: determines pulse lengths to logic unit
			Tp			=> tty_write,
			PPs 		=> tty_read,
			PPsn 		=> tty_readn,
			Ps			=> tty_Ps,
			Psn		=> tty_Psn,
			Ttrig		=> tty_trigger,
			Tread		=> tty_doread,
			Tdelay	=> tty_delay,
			Tstart	=> tty_start,
			
			-- external serial I/O
			io_clk					=> ioclk,
			uart_data_tx			=> uart_data_tx,
			uart_data_rx			=> uart_data_rx,
			write_to_uart_tx 		=> write_to_uart_tx,
			read_from_uart_rx		=> read_from_uart_rx,
			uart_rx_half_full		=> uart_rx_half_full,
			uart_rx_data_present	=> uart_rx_data_present,

			-- enable output of ESC sequences
			use_ESC					=> option(1),

			-- control panel
			readActive=> tty_reading,
			readToggle => tty_toggle,
			
			-- handshake with xmodem unit
			xmodem_active => xmodem_act,
			xmodem_done => xmodem_done,

			-- simulate typing noise via relay
			noise => tty_noise,
			
			-- debug outputs
			test			=> open
		);

	inst_xmodem: xmodem 
		port map (
			-- begin/end xmodem mode
			xmodem_active => xmodem_act,
			xmodem_done => xmodem_done,
			
			-- memory access to LGP-30 drum 
			-- system clock
			clk_tick => clk_tick,				-- can control clock from this unit via the single step mode
			
			-- main memory
			V => V,									-- read head
			Vp => Vp_xmodem,						-- write head
			W => W_xmodem, 						-- write enable
			P => P_xmodem,							-- track selector
			sec_reset => drum_sec_reset,

			A => A,
			R => R,
			C => C,
			Ap => Ap_xmodem,
			Rp => Rp_xmodem,
			Cp => Cp_xmodem,
			
			-- connection to UART unit
			io_clk					=> ioclk,
			uart_data_tx			=> uart_data_tx_xmodem,
			uart_data_rx			=> uart_data_rx,
			write_to_uart_tx 		=> write_to_uart_tx_xmodem,
			read_from_uart_rx		=> read_from_uart_rx_xmodem,
			uart_tx_half_full		=> uart_tx_half_full,
			uart_rx_half_full		=> uart_rx_half_full,
			uart_rx_data_present	=> uart_rx_data_present,

			-- debug outputs
			test => open
		);
	
	inst_uum: usb_uart_mux 
		PORT MAP (
			usb_clk		=> usbclk,
			io_clk		=> ioclk,

			-- selector input
			usb_active	=> usb_active,				-- USB data received, interface in use
			uart_active	=> uart_active,			-- UART data received, interface in use

			-- communication with internal units
			data_tx				=> uart_data_tx_mux,
			data_rx				=> uart_data_rx,
			write_strobe 		=> (write_to_uart_tx and not xmodem_act) or (write_to_uart_tx_xmodem and xmodem_act),
			read_strobe			=> (read_from_uart_rx and not xmodem_act) or (read_from_uart_rx_xmodem and xmodem_act),
			tx_half_full		=> uart_tx_half_full,
			rx_half_full		=> uart_rx_half_full,
			rx_data_present	=> uart_rx_data_present,
			
			-- external serial I/O
			uart_Rx		=> serRx,
			uart_Tx		=> serTx,
			uart_RTS		=> serRTS,
			uart_CTS		=> serCTS,

			-- external USB I/O
			usb_dp		=> usb_dp,
			usb_dn		=> usb_dn,
			usb_led		=> usb_led
		);
		
	inst_hdmi: hdmi
		PORT MAP (
			-- system clock
			clk100 => clk100_buf,
		
			-- data input
			-- only main drum memory for now
			clk => vid_clk,
			adr => vid_adr,
			data => vid_data,
			regs => vid_regs,
			drumpos => vid_drumpos,
			track => P,
			write_en => W,

			-- flip-flops
			F			=> F,
			G			=> G,
			H			=> H,
			K			=> K,
			L			=> L,
			P			=> P,
			Q			=> Q,
			
			-- display mode control
			io_clk => ioclk,
			right	=> (enc_up and not enc_sel and not enc_but),			
			left	=> (enc_dn and not enc_sel and not enc_but),			
			animate => is_slow_clock (cpow) and not xmodem_act,
			showff => option (3),
		
			-- HDMI output
			hdmi_p => hdmi_out_p,
			hdmi_n => hdmi_out_n,
			
			-- debug outputs
			test => open
		);	
		
	inst_lcd: lcd
		port map (

			-- system clock and sign bit trigger
			T	 		=> clk,
			t3 		=> t3,
			io_clk	=> ioclk,
			
			-- recycling registers
			-- Connect to the write heads, to make the display act in "real time" in single-step mode.
			-- (Different from the original LGP-30, where the oscilloscope is connected to the read heads.) 
			A		 	=> Ap,
			R			=> Rp,
			C			=> Cp,	

			-- extended diplay on?
			ext_disp => option(0),	-- toggled via option switch on Mimas board
			
			-- contrast control
			cont_up => (enc_up and enc_but),			
			cont_dn => (enc_dn and enc_but),			

			-- flip-flops
			F			=> F,
			G			=> G,
			H			=> H,
			K			=> K,
			L			=> L,
			P			=> P,
			Q			=> Q,
			
			-- clock rate display
			inv_rate	=> enc_sel,
			cpow		=> cpow,
				
			-- output to external display.
			si			=> lcd_si,
			scl		=> lcd_scl,
			a0			=> lcd_a0,
			cs_n		=> lcd_csn,
			res_n 	=> lcd_resn
		);
	
------------------------------------------------
-- global signals
------------------------------------------------
	
	
	-- multiplexed vector signals (for xmodem vs. regular operation)
	process (xmodem_act, uart_data_tx, uart_data_tx_xmodem, P, P_xmodem) is
	begin
		if xmodem_act = '1' then
			uart_data_tx_mux <= uart_data_tx_xmodem;
			P_mux <= P_xmodem;
		else
			uart_data_tx_mux <= uart_data_tx;
			P_mux <= P;
		end if;
	end process;
		
	-- status LEDs for serial connections
	usb_act_led <= usb_active;
	uart_act_led <= uart_active;

	-- Oscilloscope signals on test/expansion connector.
	-- Connect signals via resistive network, e.g.
	-- trace, fat_trace: 3.9kOhm; offset(0), offset(1): 1kOhm; sum to GND: 100pF.
	-- Note that the scope should scan right to left for correct display (low bits on the right).
	-- Turning your scope upside down will do the trick ;-)
	
	test(0)	<= fat_trace;		-- 5 MHz, can add this to traces to increase thickness
	test(2)	<= trace;			-- alternating trace, cycles through A, R, C
	test(4)	<= offset(0);		-- Y offset, bit 0
	test(6)	<= offset(1);		-- Y offset, bit 1
	test(7)	<= trig;				-- trigger goes high at the beginning of each trace 
	test(5)	<= '0' when (trig = '1') else 'Z';	-- reset sawtooth integrator

	-- other test/expansion outputs
	test(1)	<= '0';
	test(3)	<= '0';

	-- generate 5 MHz "fat trace" signal
	process (ioclk) is
	begin
		if rising_edge (ioclk) then
			fat_trace <= NOT fat_trace;
		end if;
	end process;

	-- dedicated relay output (available on landscape format PCB only)
	relay	<= tty_noise and option(2);	-- drive relay for flexowriter "typing" noise
	
end Behavioral;

