----------------------------------------------------------------------------------
--
-- Local clocks
--
-- 100 MHz input clock (Numato Mimas system clock)
--  60 MHz USB PHY layer clock
--  10 MHz system/IO clock output
-- 125 kHz LGP-30 simulation clock output 
-- 
-- jm 1.8.2016
-- jm 30.7.2017 rewritten as PLL-based clock; USB clock output added
--
----------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.numeric_std.all;
use IEEE.math_real.all;

library work;
use work.constants.all;

library unisim;
use unisim.vcomponents.all;

entity clock is
	Port ( 
		clk_ext: 	in  std_logic;				-- external clock input
		clk_usb:		out std_logic;				-- USB PHY clock, 60 MHz
		clk_io: 		out std_logic;				-- I/O clock, 10 MHz
		clk_lgp30:	out std_logic;				-- LGP-30 CPU clock. High lasts one I/O clock cycle
		clk_lgp30_unbuf: out std_logic;		-- same, but unbuffered - for use in io clock and lgp clk synchronization in panel and typewriter modules
			
		-- control for LGP-30 clock rate
		cup, cdn:	in std_logic;				-- pulse high for one clk_io cycle
		cstep:		in std_logic;				-- set high to force clock speed to single step (XModem mode)
		cpow: 		out freq_index				-- slowing clock down by 2^cdiv
	);
end clock;

architecture Behavioral of clock is

	-- PLL generated clocks (unbuffered) and control signals
   signal clk_usb_unbuf    : std_logic;
	signal clk_io_unbuf	  	: std_logic;
	signal pll_locked, clk_feedback: std_logic;

	-- derive 125 kHz clock 
	signal clk_slow: std_logic;										-- unbuffered 125 kHz clock signal
			
	signal clk_divider: natural range 0 to clk_div_max := clk_div_std;	-- current init value for clk_count
	signal clk_power: natural range cpow_min to cpow_max := cpow_std;		-- current clock speed index
	signal clk_count: natural range 0 to clk_div_max;				-- clock divide counter

begin

	-- PLL for 60MHz and 10 MHz clocks
	
   PLL_BASE_inst : PLL_BASE
   generic map (
      CLKFBOUT_MULT => 12,                  
      CLKOUT0_DIVIDE => 10,       CLKOUT0_PHASE => 0.0,   -- Output USB clock 60 MHz
      CLKOUT1_DIVIDE => 60,       CLKOUT1_PHASE => 0.0,   -- Output IO clock 10 MHz
      CLK_FEEDBACK => "CLKFBOUT",                         -- Clock source to drive CLKFBIN ("CLKFBOUT" or "CLKOUT0")
      CLKIN_PERIOD => 10.0,                              -- IMPORTANT! 100 MHz
      DIVCLK_DIVIDE => 2                                 -- Division value for all output clocks (1-52)
   )
      port map (
      CLKFBOUT => clk_feedback, 
      CLKOUT0  => clk_usb_unbuf,
      CLKOUT1  => clk_io_unbuf,
      CLKOUT2  => open,
      CLKOUT3  => open,
      CLKOUT4  => open,
      CLKOUT5  => open,
      LOCKED   => pll_locked,      
      CLKFBIN  => clk_feedback,    
      CLKIN    => clk_ext, 
      RST      => '0'              -- 1-bit input: Reset input
   );

	-- LGP-30 clock: divide 5 MHz IO clock by 40
	-- Note: Asymmetrical clock; high only for one IO clock cycle!
	-- This is used by the peripheral modules to run timers at CPU clock rate.
	
	process (clk_io_unbuf) is
	begin
		if rising_edge (clk_io_unbuf) then
		
			-- create LGP-30 clock
			if (clk_power = cpow_step) or (cstep = '1') then	-- single step mode
				clk_slow <= cdn;								-- clock ticks when encoder pulses
			else													-- regular clock
				if clk_count = 0 then                  -- counts down to divide clock
				  clk_count <= clk_divider -1;
				  clk_slow  <= '1';                    -- single cycle enable pulse
				 else
				  clk_count <= clk_count-1;
				  clk_slow <= '0';
				end if;
			end if;

			-- does the user want to change the clock rate?
			if (cup='1') and (clk_power < cpow_max) and (cstep = '0') then
				clk_divider <= freq_divider (clk_power+1);
				clk_power <= clk_power+1;
			elsif (cdn='1') and (clk_power > cpow_step)  and (cstep = '0') then
				clk_divider <= freq_divider (clk_power-1);
				clk_power <= clk_power-1;
			end if;

			-- indicate current (most recent, to be precise) clock rate
			if (cstep='0') then cpow <= clk_power;
			else cpow <= cpow_min;
			end if;
			
		end if;
	end process;
	
  -- Output buffering
 
	BUFG_ioclock  : BUFG port map ( I => clk_io_unbuf,  	O => clk_io);
   BUFG_usbclock : BUFG port map ( I => clk_usb_unbuf, 	O => clk_usb);
	BUFG_lgpclock : BUFG port map ( I => clk_slow, 			O => CLK_lgp30);
	clk_lgp30_unbuf <= clk_slow;

end Behavioral;
