----------------------------------------------------------------------------------

-- Simple encoder for LGP-30 panel.
-- Reads "raw" contacts, handles the de-bouncing locally.
-- (Only the contact used for edge detection needs de-bouncing!)
--
-- jm 3.9.16
-- jm 22.10.17  extend deadtime after each bounce

----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

library work;
use work.constants.all;


entity encoder is
	port (
		clk: 			in std_logic;
		X, Y: 		in std_logic;
		Qup, Qdn: 	out std_logic := '0'
	);
end encoder;

architecture Behavioral of encoder is

	constant deadtime: natural := 2000*CLK_IO_RATE;			-- debounce dead time 2 ms, in clk ticks
	signal deadtimer: natural range 0 to deadtime := 0;
	signal Xold: std_logic := 'Z';

begin
	process (clk)
	begin
		if rising_edge (clk) then

			Qup <= '0';								-- reset the output pulses (last 1 cycle only)
			Qdn <= '0';								-- these may be overwritten below

			if deadtimer = 0 then				-- debounce deadtime has passed
				if X='0' and Xold='1' then		-- falling edge: new encoder step
					if Y='0' then Qup <= '1';
					else Qdn <= '1';
					end if;
				end if;
			else
				deadtimer <= deadtimer -1;		-- count down debounce delay
			end if;

			-- every X transition (re-)starts the deadtime!
			-- this avoids detection of spurious falling edges due to contact bounces after the rising edge.
			if X /= Xold then deadtimer <= deadtime; end if;
			Xold <= X;

		end if;
	end process;

end Behavioral;

