----------------------------------------------------------------------------------

-- LGP-30 drum computer simulation
-- UART (used by flexowriter and xmodem)
--
-- jm 10.9.2016

-- UART from KCPSM6, Ken Chapman - Xilinx Ltd.
-- Copyright  2011-2014, Xilinx, Inc.
-- Release 9:
-- 10th May 2011 - Initial version
-- 30th July 2014 - Corrections to comments only.

-- This UART does not handle flow control.
-- RTS/CTS must be handled by the client!

----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library work;
use work.constants.all;


entity uart is
	port (
		io_clk: in std_logic;							-- I/O clock, must be 10 MHz

		-- communication with internal units
		uart_data_tx:in std_logic_vector(7 downto 0);		
		uart_data_rx: out std_logic_vector(7 downto 0);		
		write_to_uart_tx: in std_logic;				-- trigger 1 byte write to Tx buffer
		read_from_uart_rx: in std_logic;				-- trigger 1 byte read from Rx buffer
		uart_tx_half_full: out std_logic; 			-- Tx buffer >= half full?
		uart_rx_half_full: out std_logic; 			-- Rx buffer >= half full?
		uart_rx_data_present: out std_logic; 		-- Rx buffer has character?
		
		-- external serial I/O
		uart_Rx: in std_logic;
		uart_Tx: out std_logic;
		uart_RTS: out std_logic;
		uart_CTS: in std_logic
	);
end uart;


architecture Behavioral of uart is

-- UART Transmitter with integral 16 byte FIFO buffer

  component uart_tx6 
    Port (             data_in : in std_logic_vector(7 downto 0);
                  en_16_x_baud : in std_logic;
                    serial_out : out std_logic;
                  buffer_write : in std_logic;
           buffer_data_present : out std_logic;
              buffer_half_full : out std_logic;
                   buffer_full : out std_logic;
                  buffer_reset : in std_logic;
                           clk : in std_logic);
  end component;

-- UART Receiver with integral 16 byte FIFO buffer

  component uart_rx6 
    Port (           serial_in : in std_logic;
                  en_16_x_baud : in std_logic;
                      data_out : out std_logic_vector(7 downto 0);
                   buffer_read : in std_logic;
           buffer_data_present : out std_logic;
              buffer_half_full : out std_logic;
                   buffer_full : out std_logic;
                  buffer_reset : in std_logic;
                           clk : in std_logic);
  end component;

--
--
-- Signals used to connect UART_TX6
--
signal uart_tx_data_present : std_logic;
signal         uart_tx_full : std_logic;
signal 	 buf_tx_half_full  : std_logic;
signal        uart_tx_reset : std_logic := '0';	-- not used
--
-- Signals used to connect UART_RX6
--
signal         uart_rx_full : std_logic;
signal 	  buf_rx_half_full : std_logic;
signal        uart_rx_reset : std_logic := '0';		-- not used
--
-- Signals used to define baud rate
--
constant				baud_rate : natural := 19200;
constant		 baud_count_max : natural := (CLK_IO_RATE*1000000)/(16*baud_rate) -1;	-- divide counter for 16x baud_rate clock
signal           baud_count : natural range 0 to baud_count_max := 0; 
signal         en_16_x_baud : std_logic := '0';


begin

 
  -----------------------------------------------------------------------------------------
  -- UART Transmitter with integral 16 byte FIFO buffer
  -----------------------------------------------------------------------------------------
  -- Write to buffer in UART Transmitter
  
  tx: uart_tx6 
  port map (              
									data_in => uart_data_tx,
                     en_16_x_baud => en_16_x_baud,
                       serial_out => uart_tx,
                     buffer_write => write_to_uart_tx,
              buffer_data_present => uart_tx_data_present,
                 buffer_half_full => buf_tx_half_full,
                      buffer_full => uart_tx_full,
                     buffer_reset => uart_tx_reset,              
                              clk => io_clk);



  -----------------------------------------------------------------------------------------
  -- UART Receiver with integral 16 byte FIFO buffer
  -----------------------------------------------------------------------------------------
  -- Read from buffer in UART Receiver.
  -- When reading data from the receiver a pulse must be generated so that the 
  -- FIFO buffer presents the next character to be read and updates the buffer flags.
 
  rx: uart_rx6 
  port map (            serial_in => uart_rx,
                     en_16_x_baud => en_16_x_baud,
                         data_out => uart_data_rx,
                      buffer_read => read_from_uart_rx,
              buffer_data_present => uart_rx_data_present,
                 buffer_half_full => buf_rx_half_full,
                      buffer_full => uart_rx_full,
                     buffer_reset => uart_rx_reset,              
                              clk => io_clk);

 
  -----------------------------------------------------------------------------------------
  -- RS232 (UART) baud rate clock
  -----------------------------------------------------------------------------------------
  -- To set serial communication baud rate to 2400 then en_16_x_baud must pulse 
  -- High at 38,400Hz which is every 260.4 cycles at 10 MHz. 
  -- In this implementation a pulse is generated every 260 cycles.
  
   baud_clk: process(io_clk)
   begin
	 if rising_edge (io_clk) then
		if baud_count = baud_count_max then        -- counts 260 states including zero
		  baud_count <= 0;
		  en_16_x_baud <= '1';                     -- single cycle enable pulse
		 else
		  baud_count <= baud_count + 1;
		  en_16_x_baud <= '0';
		end if;
	 end if;
   end process baud_clk;
	
	
  -----------------------------------------------------------------------------------------
  -- RTS/CTS handshake 
  -----------------------------------------------------------------------------------------

	uart_RTS <= buf_rx_half_full;						--  RTS handshake output - polarity is counter-intuitive!
	uart_rx_half_full <= buf_rx_half_full;
	uart_tx_half_full <= buf_tx_half_full OR uart_CTS; -- CTS handshake input is reported as half-full buffer
	
end Behavioral;

